/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright(c) 2019 Intel Corporation. All rights reserved. */

#ifndef __ACCFG_H__
#define __ACCFG_H__

#ifndef ARRAY_SIZE
#include <ccan/array_size/array_size.h>
#endif
#include <linux/types.h>
#include <sys/user.h>
#include <unistd.h>

#endif /* __ACCFG_H__ */
